# -*- coding: utf-8 -*-
#
# This file is a part of the CaosDB Project.
#
# Copyright (C) 2023 Henrik tom Wörden <h.tomwoerden@indiscale.com>
# Copyright (C) 2023 IndiScale GmbH <info@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

"""
This module provides some cached versions of functions that retrieve Entities from a remote server.
"""

from typing import Union
from functools import lru_cache
from .utils import get_entity
from .common.models import execute_query, Entity, Container


from enum import Enum

# roughly 1GB for typical entity sizes
DEFAULT_SIZE = 33333


# Those dict caches are solely for filling the real cache manually (e.g. to reuse older query
# results)
_DUMB_BY_NAME_CACHE = {}
_DUMB_BY_PATH_CACHE = {}
_DUMB_BY_EID_CACHE = {}
_DUMB_BY_QUERY_CACHE = {}


class AccessType(Enum):
    """ This module looks for entities based on those kinds of information. """
    QUERY = 1
    PATH = 2
    EID = 3
    NAME = 4


def cached_get_entity_by(eid: Union[str, int] = None, name: str = None, path: str = None, query:
                         str = None) -> Entity:
    """ returns a single entity that is identified uniquely by one of the arguments

    You must supply exactly one of the arguments.
    """
    count = 0
    if eid is not None:
        count += 1
    if name is not None:
        count += 1
    if path is not None:
        count += 1
    if query is not None:
        count += 1
    if count != 1:
        raise ValueError("You must supply exactly one argument.")

    if eid is not None:
        return _cached_access(AccessType.EID, eid, unique=True)
    if name is not None:
        return _cached_access(AccessType.NAME, name, unique=True)
    if path is not None:
        return _cached_access(AccessType.PATH, path, unique=True)
    if query is not None:
        return _cached_access(AccessType.QUERY, query, unique=True)


def cached_query(query_string) -> Container:
    """ a cached version of db.execute_query """
    return _cached_access(AccessType.QUERY, query_string, unique=False)


@lru_cache(maxsize=DEFAULT_SIZE)
def _cached_access(kind: AccessType, value: Union[str, int], unique=True):
    # This is the function that is actually cached.
    # Due to the arguments, the cache has kind of separate sections for cached_query and
    # cached_get_entity_by with the different AccessTypes. However, there is only one cache size.

    # The dumb dict caches are only to allow filling the cache manually
    if kind == AccessType.QUERY:
        if value in _DUMB_BY_QUERY_CACHE:
            return _DUMB_BY_QUERY_CACHE[value]
        return execute_query(value)
    elif kind == AccessType.NAME:
        if value in _DUMB_BY_NAME_CACHE:
            return _DUMB_BY_NAME_CACHE[value]
        return get_entity.get_entity_by_name(value)
    elif kind == AccessType.EID:
        if value in _DUMB_BY_EID_CACHE:
            return _DUMB_BY_EID_CACHE[value]
        return get_entity.get_entity_by_id(value)
    elif kind == AccessType.PATH:
        if value in _DUMB_BY_PATH_CACHE:
            return _DUMB_BY_PATH_CACHE[value]
        return get_entity.get_entity_by_path(value)


def cache_clear() -> None:
    """ Empty the cache that is used by cached_query and cached_get_entity_by """
    _cached_access.cache_clear()


def cache_info():
    """ Empty the cache that is used by cached_query and cached_get_entity_by """
    return _cached_access.cache_info()


def cache_initialize(maxsize=DEFAULT_SIZE) -> None:
    """ Create a new cache with the given size for cached_query and cached_get_entity_by

    The old cache is removed with all its content.
    """
    global _cached_access
    _cached_access = lru_cache(maxsize=maxsize)(_cached_access.__wrapped__)


def fill_cache(items: dict, kind: AccessType = AccessType.EID, unique=True) -> None:
    """ add Entities to the cache manually

    This allows to fill the cache without actually submitting queries.
    """
    # 1. add the given items to the corresponding dumb dict cache
    if kind == AccessType.EID:
        _DUMB_BY_EID_CACHE.update(items)
    elif kind == AccessType.NAME:
        _DUMB_BY_NAME_CACHE.update(items)
    elif kind == AccessType.PATH:
        _DUMB_BY_PATH_CACHE.update(items)
    elif kind == AccessType.QUERY:
        _DUMB_BY_QUERY_CACHE.update(items)

    # 2. call the cache function with each key (this only results in a dict look up)
    for key in items.keys():
        _cached_access(kind, key, unique=unique)

    # 3. empty the dumpy dict cache again
    if kind == AccessType.EID:
        _DUMB_BY_EID_CACHE.clear()
    elif kind == AccessType.NAME:
        _DUMB_BY_NAME_CACHE.clear()
    elif kind == AccessType.PATH:
        _DUMB_BY_PATH_CACHE.clear()
    elif kind == AccessType.QUERY:
        _DUMB_BY_QUERY_CACHE.clear()
